# Spectral.jl

using DSP
using FFTW

# PRIVATE FUNCTION
#
# The purpose of this function is to add a segment to the buffer used for 
# computing spectral information for the signal.  The function updates the 
# complex-valued buffer XFRM with the FFT of XNEW.
function add_segment!(
    window::Function,
    winsize::Int,
    xnew::Array{T},
    xfrm::Array{S},
    firsttime::Bool
) where {T <: Real, S <: Complex}
    # Fill the buffer with windowed data
    w = window(winsize)
    sumw = sum(w)
    factor = winsize / sumw
    buffer = w .* xnew[1:winsize]

    # Compute & scale the FFT
    p = rfft(buffer)
    m = length(p)
    p = p ./ m

    # Update the buffer
    if firsttime
        for i in 1:m
            amp = factor * abs(p[i])
            phase = atan(imag(p[i]), real(p[i]))
            xfrm[i] = complex(amp * cos(phase), amp * sin(phase))
        end
    else
        for i in 1:m
            amp = factor * abs(p[i])
            # phase = atan(imag(p[i]), real(p[i]))

            amp2 = abs(xfrm[i])
            phase = atan(imag(xfrm[i]), real(xfrm[i]))

            xfrm[i] = complex(
                (amp * cos(phase) + amp2 * cos(phase)),
                (amp * sin(phase) + amp2 * sin(phase))
            )
        end
    end

    # End
    return nothing
end

# ------------------------------------------------------------------------------
# PRIVATE FUNCTION
#
# The purpose of this function is to overlap windowed Fourier transforms of the 
# supplied data set in XNEW.
function overlap_segment(
    window::Function,
    winsize::Int,
    xnew::Array{T},
    xfrm::Array{S},
    firsttime::Bool,
    counter::Int
) where {T <: Real, S <: Complex}
    # Determine size information
    nt = length(xnew)
    m = compute_xfrm_length(winsize)
    nk = div(nt - 1, m)
    del = nk > 1 ? (nt - winsize) / (nk - 1.0) : 0.0

    # Process
    c = counter
    buffer = zeros(T, winsize)
    if nt < winsize
        # Pad with zeros
        buffer[1:nt] = xnew
        add_segment!(window, winsize, buffer, xfrm, firsttime)
        c = c + 1
    else
        # Overlap
        for k in 0:nk-1
            noff = floor(Int64, k * del + 0.5)
            nend = noff + winsize
            buffer = xnew[noff + 1:nend]
            add_segment!(window, winsize, buffer, xfrm, firsttime)
            c = c + 1
            firsttime = false
        end
    end

    # Output
    return c
end

# ------------------------------------------------------------------------------
# PRIVATE FUNCTION
#
# Computes the windowed Fourier transform of the supplied data set.  Only the 
# positive half of the discrete transform is returned.
function spectrum(
    window::Function,
    winsize::Int,
    x::Array{T}
) where T <: Real
    # Create a complex-valued array to store the transformed data
    npts = compute_xfrm_length(winsize)
    xfrm = zeros(ComplexF64, npts)

    # Compute the transform
    count = overlap_segment(window, winsize, x, xfrm, true, 0)

    # Average the results
    if count > 1
        nt = length(xfrm)
        for i in 1:nt
            amp = abs(xfrm[i]) / count
            phase = atan(imag(xfrm[i]), real(xfrm[i]))
            xfrm[i] = complex(amp * cos(phase), amp * sin(phase))
        end
    end

    # End
    return xfrm
end

# ------------------------------------------------------------------------------
# PRIVATE FUNCTION
#
# Computes the proper length of a discrete Fourier transform.
function compute_xfrm_length(n::Int)
    # return mod(n, 2) == 0 ? n / 2 + 1 : (n + 1) / 2
    return mod(n, 2) == 0 ? div(n, 2) + 1 : div(n + 1, 2)
end
