using FrequencySweep
using Test
using DSP
using Plots

@testset "FrequencySweep.jl" begin
    @testset "Spectral" begin
        # ----------------------------------------------------------------------
        # COMPUTE_XFRM_LENGTH
        n1 = FrequencySweep.compute_xfrm_length(500)
        @test isequal(n1, 251)

        n2 = FrequencySweep.compute_xfrm_length(501)
        @test isequal(n2, 251)

        # ----------------------------------------------------------------------
        # ADD_SEGMENT! - FIRST TIME TEST
        t = 0:1e-3:1.0
        X = 2.5
        freq = 50.0
        x = X .* sin.((2 * pi * freq) .* t)

        xfrm = zeros(
            ComplexF64,
            floor(Int64, FrequencySweep.compute_xfrm_length(length(t)))
        )
        FrequencySweep.add_segment!(
            hanning,
            length(xfrm),
            x,
            xfrm,
            true
        )

        tol = 0.02
        maxX = maximum(abs.(xfrm))
        @test isless(abs(X - maxX), tol)

        # ----------------------------------------------------------------------
        # ADD_SEGMENT! - SECOND TIME TEST
        FrequencySweep.add_segment!(
            hanning,
            length(xfrm),
            x,
            xfrm,
            false
        )

        maxX1a = maximum(abs.(xfrm))
        @test isless(abs(2 * X - maxX1a), tol)

        # ----------------------------------------------------------------------
        # OVERLAP_SEGMENT - OVERLAP LENGTH TEST
        xfrm2 = zeros(
            ComplexF64,
            500
        )
        count = FrequencySweep.overlap_segment(
            hanning,
            length(xfrm2),
            x,
            xfrm2,
            true,
            0
        )

        maxX2 = maximum(abs.(xfrm2))
        @test isless(abs(X - maxX2 / count), tol)

        # ----------------------------------------------------------------------
        # OVERLAP_SEGMENT - PAD WITH ZEROS TEST
        xfrm3 = zeros(
            ComplexF64,
            length(t)
        )
        count = FrequencySweep.overlap_segment(
            hanning,
            length(xfrm3),
            x,
            xfrm3,
            true,
            0
        )
        xfrm3 = xfrm3 ./ count  # need to average

        @test isequal(count, 1) # only 1 time through for a pad-with-zeros

        maxX3 = maximum(abs.(xfrm3))
        @test isless(abs(X - maxX3), tol)

        # ----------------------------------------------------------------------
        # SPECTRUM - OVERLAP TEST
        X1 = FrequencySweep.spectrum(
            hanning,
            500,
            x
        )
        check1 = maximum(abs.(X1))
        @test isless(abs(check1 - X), tol)
    end

    @testset "Sweep" begin
        # Duffing Model Parameters
        # x" + delta * x' + alpha * x + beta * x^3 = gamma * sin(omega * t)
        alpha = 1.0
        gamma = 1.0
        delta = 0.1
        beta = 0.05

        # Duffing's Equation
        function duffing_eqn(du, u, p, omega, t)
            du[1] = u[2]
            du[2] = gamma * sin(omega * t) - delta * u[2] - alpha * u[1] - beta * u[1]^3
        end
        # Analytical Solution - computes frequency values given an amplitude
        function leg1(z)
            arg = 4 * gamma^2 - 3 * beta * delta^2 * z^4 + 
                (delta^2 - 4 * alpha) * delta^2 * z^2
            s = arg < 0.0 ? 
                NaN : 
                (2 * sqrt(arg) + z * (3 * beta * z^2 - 2 * delta^2 + 4 * alpha)) / (4 * z)
            return s < 0.0 ? NaN : sqrt(s)
        end
        function leg2(z)
            arg = 4 * gamma^2 - 3 * beta * delta^2 * z^4 + 
                (delta^2 - 4 * alpha) * delta^2 * z^2
            s = arg < 0.0 ?
                NaN :
                (-2 * sqrt(arg) + z * (3 * beta * z^2 - 2 * delta^2 + 4 * alpha)) / (4 * z)
            return s < 0.0 ? NaN : sqrt(s)
        end

        # Plot both legs of the actual solution
        z = 0.5:0.05:10.0
        magplt = plot(
            leg1.(z), z,
            label = "Analytical",
            lw = 2,
            lc = :steelblue,
            xlabel = "\u03C9",
            ylabel = "X",
            legend = :topleft
        )
        plot!(
            magplt,
            leg2.(z), z,
            label = false,
            lw = 2,
            lc = :steelblue
        )

        # Frequency Points
        f1 = Array(0.5:0.05:2.0)
        f2 = Array(2.0:-0.05:0.5)

        # Perform the sweep
        sweep1 = frf_sweep(
            duffing_eqn,
            [0.0, 0.0],
            f1
        )
        sweep2 = frf_sweep(
            duffing_eqn,
            [0.0, 0.0],
            f2
        )

        # Plot the results
        plot!(
            magplt,
            f1, abs.(sweep1[:,1]),
            label = "Ascending",
            markershape = :circle,
            lc = :green,
            markercolor = :green
        )
        plot!(
            magplt,
            f2, abs.(sweep2[:,1]),
            label = "Descending",
            markershape = :xcross,
            lc = :orange,
            markercolor = :orange
        )
        
        phaseplt = plot(
            f1, angle.(sweep1[:,1]),
            label = false,
            markershape = :circle,
            xlabel = "\u03C9",
            ylabel = "\u03D5",
            lc = :green,
            markercolor = :green
        )
        plot!(
            f2, angle.(sweep2[:,1]),
            label = false,
            markershape = :xcross,
            lc = :orange,
            markercolor = :orange
        )

        plt = plot(
            magplt, phaseplt,
            layout = (2, 1)
        )
        savefig(plt, "Duffing.png")
    end
end
